import sys
import os

FILENAME = "qtrvsim_tester.py"
SRC_DIR = os.path.realpath(__file__).replace(FILENAME, "")

sys.path.append(SRC_DIR+"code")
import constants as cn
import myparse as mp
import testing as ts
import selftesting as sts

parser = mp.init_parser()
params = parser.parse_args()
if (params.dregs):
    params.nodump = 0
    params.nopass = 0

if(params.clean):
    ts.delete_elf(SRC_DIR)
    sys.exit(0)
        
sim_bin, bin_check = ts.test_sim_bin(params.qtrvsim_cli)

if(not bin_check):
    print("Problem with qtrvsim_cli binary!")
    sys.exit(1)

self_files, s_file_check = ts.load_filenames(
    SRC_DIR + cn.SELF_PATH, bool(params.rebuild))
test_files, t_file_check = ts.load_filenames(
    SRC_DIR + cn.ISA_PATH, bool(params.rebuild))

if (params.selftest):
    sts.self_test(sim_bin, params, SRC_DIR, self_files)

ts.test_selector(sim_bin, params, SRC_DIR, test_files)

sys.exit(0)



import os
import sys
import pathlib
import subprocess

import helpers as hp
import constants as cn


def test_sim_bin(sim_bin):
    try:
        sim_test = subprocess.run(
            [sim_bin, "-h"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except:
        parser.print_help()
        sys.exit(1)
    if (sim_test.returncode == 0):
        return sim_bin, True
    return "", False


def load_filenames(dir_path, rebuild):
    test_files = list(filter(
        lambda x: not str(x).startswith("."), os.listdir(dir_path+"/elf")))
    if (len(test_files) != 0 and not rebuild):
        return test_files, True
    else:
        try:
            if (rebuild):
                print("Building {0}".format(dir_path))
            else:
                print("No tests found. Trying to build tests.")
            tests_built = subprocess.run(
                ["make", "-C", dir_path],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL)
            return load_filenames(dir_path, False)
        except:
            print("Failed to build tests. {0} ; {1}".format(
                dir_path, test_files))
            sys.exit(1)


# If output of the qtrvsim-cli is changed this will probably break !!!
def check_reg_dump(reg_dump):
    res = 2
    stdout = str(reg_dump.stdout)
    if (stdout.__contains__("R11:")):
        index = stdout.find("R11:")
        if (stdout[index:index+18].__contains__("600d")):
            res = 1
        elif (stdout[index:index+18].__contains__("bad")):
            res = 0
    return res, reg_dump


def run_test(sim_bin, params, dir_path, filename):
    test_path = dir_path + filename
    try:
        param_bin = [sim_bin, test_path, "--d-regs"]
        if (params.pipeline):
            param_bin.append("--pipelined")
        if (params.cache):
            param_bin.extend(cn.CACHE_SETTINGS)
        return subprocess.run(param_bin, capture_output=True)
    except subprocess.CalledProcessError as err:
        print(err)
        sys.exit(1)


def run_tests(sim_bin, params, dir_path, tests):
    succ = 0
    max_width = 0
    if (not params.fileprt):
        max_width = hp.max_str_list(tests)
    for test in tests:
        reg_dump = run_test(sim_bin, params, dir_path, test)
        test = test.ljust(max_width)
        test_res, test_reg_dump = check_reg_dump(reg_dump)
        succ += hp.res_print(test, test_res, test_reg_dump, params)
    if (not params.fileprt):
        print(str(succ) + "/" + str(len(tests)) + " tests succesfull.\n")


def run_official_tests(sim_bin, params, src_path, tests):
    if (not params.no32):
        if (not params.fileprt):
            print("--- 32 bit register tests ---")
        run_tests(sim_bin, params, src_path + cn.ELF_PATH, tests[0])
    if (not params.no64):
        if (not params.fileprt):
            print("--- 64 bit register tests ---")
        run_tests(sim_bin, params, src_path + cn.ELF_PATH, tests[1])


def test_selector(sim_bin, params, src_path, tests):
    if (params.pipeline):
        print("Simulator runs in pipelined mode.")
    if (params.cache):
        print("Simulator runs in cache mode.")
    if (params.fileprt):
        line = ""
    else:
        line = "-+-+-+-+-+-+-+-+-"
    print(line+"RVxxUI"+line)
    run_official_tests(sim_bin, params, src_path, hp.get_RVxx(tests, "ui"))

    if (params.multiply):
        print(line+"RVxxUM"+line)
        run_official_tests(sim_bin, params, src_path, hp.get_RVxx(tests, "um"))

    if (params.atomic):
        print(line+"RVxxUA"+line)
        run_official_tests(sim_bin, params, src_path, hp.get_RVxx(tests, "ua"))

    if (params.CSR):
        print(line+"RVxxSI"+line)
        run_official_tests(sim_bin, params, src_path, hp.get_RVxx(tests, "si"))
        print(line+"RVxxMI"+line)
        run_official_tests(sim_bin, params, src_path, hp.get_RVxx(tests, "mi"))

    if (len(params.external) > 0):
        print(line+"External Tests"+line)
        dir_path = params.external[0]
        test_files = os.listdir(dir_path)
        test_files = list(
            filter(lambda x: not str(x).startswith("."), test_files))
        run_tests(sim_bin, params, dir_path, test_files)


def delete_elf(src_path):
    try:
        subprocess.run(["make", "-C", src_path + cn.ISA_PATH, "clean"],
                       stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL)
        subprocess.run(["make", "-C", src_path + cn.SELF_PATH, "clean"],
                       stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError as err:
        print("Unable to delete elf files.")
        print(err)



import testing as ts
import constants as cn

def self_test(sim_bin, params, src_path, tests):
    sh_pass = []
    sh_fail = []
    error = []
    for test in tests:
        test_res, test_reg_dump = ts.check_reg_dump(
            ts.run_test(sim_bin, params, src_path + cn.SELF_ELF_PATH+"/", test))
        if (test_res == 1 and str(test).__contains__("fail")):   
            sh_fail.append(test)
        if (test_res == 0 and str(test).__contains__("pass")):   
            sh_pass.append(test)
        if(test_res == 2):                                        
            error.append(test)
    print("Test passed but was expected to fail:{0}".format(sh_fail))
    print("Test failed but was expected to pass:{0}".format(sh_pass))
    print("Test ended in error                 :{0}\n".format(error))


import argparse


def init_parser():
    parser = argparse.ArgumentParser(
        description="Defeault tests are RV32UI and RV64UI. No pipeline and cache.")
    parser.add_argument('qtrvsim_cli',
                        default='',
                        help="Qtrvsim_cli to run tests. (RVxxUI is default)")

    # Test selection
    bsel = parser.add_argument_group(
        "bit lenght selection for official tests:")
    bsel.add_argument("--no-32",
                      action="count",
                      default=0,
                      dest="no32",

                      help="Disables 32-bit tests. (Only for official tests)")
    bsel.add_argument("--no-64",
                      action="count",
                      default=0,
                      dest="no64",
                      help="Disables 64-bit tests. (Only for official tests)")

    # Additional tests
    tsel = parser.add_argument_group("additional set of tests:")
    tsel.add_argument("-E", "--external",
                      nargs=1,
                      action="store",
                      default="",
                      help="Path to additional tests. Required to adhere to PASS,FAIL behavior of edited test enviroment.")
    tsel.add_argument("-M", "--multiply",
                      action="count",
                      default=0,
                      help="Additional set of tests for multiplication.")
    tsel.add_argument("-A", "--atomic",
                      action="count",
                      default=0,
                      help="Additional set of tests for atomic instructions.")
    tsel.add_argument("--CSR",
                      action="count",
                      default=0,
                      help="Additional set of tests for control and status registers.")

    # QtRVSim_cli settings
    opts = parser.add_argument_group("QtRVSim options:")
    opts.add_argument("--pipeline",
                      action="count",
                      default=0,
                      help="Simulator runs in pipelined configuration.")
    opts.add_argument("--cache",
                      action="count",
                      default=0,
                      help="Simulator runs with d-cache and i-cache implemented. (d lru,2,2,2,wb ; i lru,2,2,2)")

    # Test aplication output options
    eout = parser.add_argument_group("output print options:")
    eout.add_argument("--no-pass",
                      action="count",
                      default=0,
                      dest="nopass",
                      help="Prints only failed tests for readability.")
    eout.add_argument("--no-dump",
                      action="count",
                      default=0,
                      dest="nodump",
                      help="Omits printing of registers in failed test.")
    eout.add_argument("--d-regs",
                      action="count",
                      default=0,
                      dest="dregs",
                      help="Prints output of qtrvsim --d-regs param. (Higher priority than other output options.)")
    eout.add_argument("--file",
                      action="count",
                      default=0,
                      dest="fileprt",
                      help="Removes most graphical output features.")

    # Test aplication good to have features
    gth = parser.add_argument_group("good to have:")
    gth.add_argument("-R", "--rebuild",
                     action="count",
                     default=0,
                     help="Deletes tests and buidls them anew.")
    gth.add_argument("-S", "--selftest",
                     action="count",
                     default=0,
                     dest="selftest",
                     help="Enable simple tests to check if tester behaves properly.")
    gth.add_argument("--clean",
                     action="count",
                     default=0,
                     dest="clean",
                     help="Deletes all official and seftest elf files.")
    return parser



def max_str_list(list):
    if (len(list) > 0):
        return len(max(list, key=lambda x: len(x)))
    return 0


# Decorative method, if output of the qtrvsim-cli is changed this will probably break!!!
def print_formated_output(reg_dump):
    stdout = reg_dump.stdout.decode("utf-8")
    stderr = reg_dump.stderr.decode("utf-8")
    print(stderr)
    print(stdout)


def res_translate(res):
    if (res == 0):
        return "FAIL"
    elif (res == 1):
        return "PASS"
    else:
        return "ERROR"


def res_print(test, test_res, test_reg_dump, params):
    if (test_res == 1):
        if (not params.nopass):
            if (not params.fileprt):
                print(test + ": " + '\033[92m' + "PASS" + '\033[0m')
            else:
                print(test+":PASS")
            if (params.dregs):
                print_formated_output(test_reg_dump)
        return 1
    elif (test_res == 0):
        if (not params.fileprt):
            print(test + ": " + '\033[93m' + "FAIL" + '\033[0m')
        else:
            print(test+":FAIL")
        if (not params.nodump or params.dregs):
            print_formated_output(test_reg_dump)
        return 0
    else:
        if (not params.fileprt):
            print(test + ": " + '\033[91m' + "ERROR" + '\033[0m')
        else:
            print(test+":ERROR")
        if (not params.nodump or params.dregs):
            print_formated_output(test_reg_dump)
        return 0


def get_RVxx(tests, isa):
    isa = str(isa).lower()
    tests32 = list(filter(lambda t: str(t).__contains__("32"+isa), tests))
    tests64 = list(filter(lambda t: str(t).__contains__("64"+isa), tests))
    tests32.sort()
    tests64.sort()
    return tests32, tests64